#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

static char *verstr=
  "MoDapple  Version 0.20\n"
  "Copyright 2016 Steve Nickolas\n"
  "Other copyrights apply to plugin code\n";
static char *ESTR_NOTMOD ="%s is not a module\n",
            *ESTR_NOFUNC ="Could not find a '%s' function in %s\n",
            *ESTR_FSIZE  ="%s: Wrong file size (%lu expected, %lu found)\n",
            *ESTR_SHORT  ="%s: Short read (%lu expected, %lu found\n",
            *ESTR_INITMOD="Could not initialize module %s\n",
            *ESTR_SLOT   ="Slot %d",
            *ESTR_IS     =" is %s Version %s\n%s\n",
            *ESTR_LINK   ="Could not link %s to slot %d\n",
            *ESTR_NOTTY  ="No console loaded\n",
            *ESTR_NOCPU  ="No CPU loaded\n",
            *ESTR_NOBIOS ="No BIOS loaded\n",
            *ESTR_BADSLOT="Invalid slot %d\n",
            *ESTR_OPEN   ="Could not open module %s\n",
            *ESTR_ISNOT  ="%s is not a ",
            *ESTR_NOTSLOT="slot module\n",
            *ESTR_NOTTTY ="console module\n",
            *ESTR_NOTCPU ="CPU module\n",
            *ESTR_CPU    ="CPU",
            *ESTR_TTY    ="Console";

static char *STRING_UNDERSCORE_STRING="%s_%s";
static char *MOD_TYPE="mod_type",
            *MOD_NAME="mod_name",
            *MOD_PREF="mod_pref",
            *MOD_VERS="mod_vers",
            *MOD_COPR="mod_copr",
            *INI_SYSTEM   ="System",
            *INI_SLOTCARDS="SlotCards",
            *INI_BIOS     ="BIOS",
            *INI_CONSOLE  ="Console",
            *INI_CPU      ="CPU",
            *INI_DEF_ROM  ="apple2p.rom";

static char *ARBY="rb"; /* for reading files */

typedef struct
{
 HMODULE  dll;
 uint32_t *mod_type;
 uint8_t  **mod_name;
 uint8_t  **mod_pref;
 uint8_t  **mod_vers;
 uint8_t  **mod_copr;
 uint8_t  **preload_name;
 uint16_t *preload_size;
 uint8_t  **preload_ptr;
 void     (*write16)     (uint16_t a, uint8_t v);
 void     (*write256)    (uint16_t a, uint8_t v);
 void     (*write2048)   (uint16_t a, uint8_t v);
 uint8_t  (*read16)      (uint16_t a);
 uint8_t  (*read256)     (uint16_t a);
 uint8_t  (*read2048)    (uint16_t a);
 int16_t  (*override_rd) (uint16_t a);
 int16_t  (*override_wr) (uint16_t a, uint8_t v);
 int      (*reset)       (void);
 int      (*loop)        (int cycles);
 
 /*
  * Formerly, init() was a boolean.
  * For plugins intended for MoDapple 0.09 and later, if init() returns "6502"
  * then it needs pointers to cpuread() and cpuwrite().  The function to take
  * these pointers is only linked if it is required.
  */
 int      (*init)        (void);
 int      (*deinit)      (void);
 
 void     (*dmainit)     (void (*v_cpuwrite)(uint16_t, uint8_t), uint8_t (*v_cpuread)(uint16_t));

 void     (*config)      (void);
} SLOT;

SLOT slot[8];

HINSTANCE hinstance;

typedef struct {
 uint8_t iiemode; /* emulating a //e */
 uint8_t store80,
         rdaux,
         wraux,
         lcaux,
         cxoverride,
         c3override,
         txt80,
         dblgrx,
         mostxt,
         iou,
         an3,
         alt[2];
 uint8_t lcread,
         lcwrite,
         lcalt;
 uint8_t nogx,
         split,
         page2,
         hires;
} IIE;

IIE iie;

uint8_t  ram[131072], rom[16384];

uint8_t  fnstr[256];

/* Data pointers and functinos for TTY driver */
uint32_t *tty_mod_type;
uint8_t  **tty_mod_name;
uint8_t  **tty_mod_pref;
uint8_t  **tty_mod_vers;
uint8_t  **tty_mod_copr;
uint8_t  *ttykey;
void     (*ttywrite)   (uint32_t a, uint8_t v);
void     (*ttyiie)     (void);
void     (*ttyout)     (uint8_t a, uint8_t v);
uint8_t  (*ttyin)      (uint8_t a);
void     (*ttydeinit)  (void);
void     (*ttyinit)    (IIE*);
int      (*ttyloop)    (void);

/* Data pointers and functinos for CPU driver */
uint32_t *cpu_mod_type;
uint8_t  **cpu_mod_name;
uint8_t  **cpu_mod_pref;
uint8_t  **cpu_mod_vers;
uint8_t  **cpu_mod_copr;
void     (*cpureset)     (void);
void     (*cpurun)       (void);
void     (*cpuinit)      (void (*v_cpuwrite)(uint16_t, uint8_t), uint8_t (*v_cpuread)(uint16_t), int8_t (*v_cpuloop)(uint32_t));
void     (*cpuseta)      (uint8_t val);
void     (*cpusetx)      (uint8_t val);
void     (*cpusety)      (uint8_t val);
void     (*cpusets)      (uint8_t val);
void     (*cpusetp)      (uint8_t val);
void     (*cpusetpc)     (uint16_t val);
void     (*cpusetrate)   (int val);
uint8_t  (*cpugeta)      (void);
uint8_t  (*cpugetx)      (void);
uint8_t  (*cpugety)      (void);
uint8_t  (*cpugets)      (void);
uint8_t  (*cpugetp)      (void);
uint16_t (*cpugetpc)     (void);
int      (*cpugetrate)   (void);
int      (*cpugetcycles) (void);

HMODULE mod_tty, mod_cpu;

void canlibs (void)
{
 if (mod_tty) FreeLibrary(mod_tty);
 if (mod_cpu) FreeLibrary(mod_cpu);
}

void notamod (char *filename)
{
 canlibs();
 fprintf (stderr, ESTR_ISNOT, filename);
 fprintf (stderr, ESTR_NOTMOD);
 exit(1);
}

void nofunc (char *filename, char *ident)
{
 canlibs();
 fprintf (stderr, ESTR_NOFUNC, ident, filename);
 exit(1);
}

int nofunc2 (int slotno, char *filename, char *ident)
{
 FreeLibrary(slot[slotno].dll);
 slot[slotno].dll=0;
 fprintf (stderr, ESTR_NOFUNC, ident, filename);
 return -1;
}

#define TFUNCTINO(a,b) sprintf (fnstr,STRING_UNDERSCORE_STRING,*tty_mod_pref,b); if (!(a=(void*)GetProcAddress(mod_tty,fnstr))) nofunc(ttymod, b)
#define CFUNCTINO(a,b) sprintf (fnstr,STRING_UNDERSCORE_STRING,*cpu_mod_pref,b); if (!(a=(void*)GetProcAddress(mod_cpu,fnstr))) nofunc(cpumod, b)
#define SFUNCTINO(s,a,b) sprintf (fnstr,STRING_UNDERSCORE_STRING,*(slot[s].mod_pref),b); if (!(slot[s].a=(void*)GetProcAddress(slot[s].dll,fnstr))) return nofunc2(s, module, b)

int bload (char *filename, int expect, uint8_t *pointer, uint8_t quiet)
{
 FILE *file;
 long l;
 
 file=fopen(filename,ARBY);
 if (!file)
 {
  perror(filename);
  return -1;
 }
 
 fseek(file,0,SEEK_END);
 l=ftell(file);
 fseek(file,0,SEEK_SET);
 if (l!=expect)
 {
  fclose(file);
  if (!quiet) fprintf (stderr, ESTR_FSIZE, filename, expect, l);
  return -1;
 }
 l=fread(pointer,1,expect,file);
 fclose(file);
 if (l<expect)
 {
  fprintf(stderr, ESTR_SHORT, filename, expect, l);
  return -1;
 }
 
 return 0;
}

/*****************************************************************************/
#if 0
/*
 * Start of a joystick driver.  XXX: This code should be in a DLL.
 * XXX: How is the xy info sent to the ][?
 */

static JOYINFO joyinfo;
static unsigned int nojoys, whichjoy;
static BOOL idjoy1, idjoy2;
static int gotstick;

typedef struct
{
 int8_t x;
 int8_t y;
 uint8_t button[2];
} AIIJOY;

static AIIJOY aiijoy;

void joyread (void)
{
 joyGetPos(whichjoy,&joyinfo);
 
 aiijoy.x=(joyinfo.wXpos>>8)-127;
 aiijoy.y=(joyinfo.wYpos>>8)-127;
 aiijoy.button[0]=(joyinfo.wButtons&JOY_BUTTON1)?0x00:0x80;
 aiijoy.button[1]=(joyinfo.wButtons&JOY_BUTTON2)?0x00:0x80;
}

void joyinit (void)
{
 gotstick=0;
 
 if (!(nojoys=joyGetNumDevs())) return;
 
 idjoy1=(joyGetPos(JOYSTICKID1,&joyinfo)!=JOYERR_UNPLUGGED);
 idjoy2=((nojoys==2)&&(joyGetPos(JOYSTICKID2,&joyinfo)!=JOYERR_UNPLUGGED));
 if (idjoy1||idjoy2)
 {
  gotstick=1;
  whichjoy=idjoy1?JOYSTICKID1:JOYSTICKID2;
 }
}
#endif
/*****************************************************************************/

static int fast6, fast6on;

static int holdc8;

/*
 * The ][+'s language card is a plugin, so that it can be added and removed at
 * will.  The //e and later MUST have this feature, and so a second language
 * card is hardwired into the shell for this support.
 */

uint8_t lclast;

static void iielc (uint16_t a)
{
 switch (a&0x0F)
 {
  case 0x0:
  case 0x4:
   iie.lcalt=0x80;
   iie.lcread=0x80;
   iie.lcwrite=0x00;
   break;
  case 0x1:
  case 0x5:
   iie.lcalt=0x80;
   iie.lcread=0x00;
   iie.lcwrite=0x80;
   break;
  case 0x2:
  case 0x6:
   iie.lcalt=0x80;
   iie.lcread=0x00;
   iie.lcwrite=0x00;
   break;
  case 0x3:
  case 0x7:
   iie.lcalt=0x80;
   iie.lcread=0x80;
   iie.lcwrite=0x80;
   break;
   
  case 0x8:
  case 0xC:
   iie.lcalt=0x00;
   iie.lcread=0x80;
   iie.lcwrite=0x00;
   break;
  case 0x9:
  case 0xD:
   iie.lcalt=0x00;
   iie.lcread=0x00;
   iie.lcwrite=0x80;
   break;
  case 0xA:
  case 0xE:
   iie.lcalt=0x00;
   iie.lcread=0x00;
   iie.lcwrite=0x00;
   break;
  case 0xB:
  case 0xF:
   iie.lcalt=0x00;
   iie.lcread=0x80;
   iie.lcwrite=0x80;
   break;
 }
 
 if (lclast!=(a&0x0F)) iie.lcwrite=0x00;
 lclast=(a&0x0F);
}

void cpuwrite(uint16_t a, uint8_t v)
{
 int t, o;
 
 if ((iie.iiemode)&&((a&0xFFF0)==0xC000))
 {
  switch (a&0x000F)
  {
   case 0x0:
    iie.store80=0x00;
    break;
   case 0x1:
    iie.store80=0x80;
    break;
   case 0x2:
    iie.rdaux=0x00;
    break;
   case 0x3:
    iie.rdaux=0x80;
    break;
   case 0x4:
    iie.wraux=0x00;
    break;
   case 0x5:
    iie.wraux=0x80;
    break;
   case 0x6:
    iie.cxoverride=0x00;
    break;
   case 0x7:
    iie.cxoverride=0x80;
    break;
   case 0x8:
    iie.lcaux=0x00;
    break;
   case 0x9:
    iie.lcaux=0x80;
    break;
   case 0xA:
    iie.c3override=0x00;
    break;
   case 0xB:
    iie.c3override=0x80;
    break;
   case 0xC:
    iie.txt80=0x00;
    break;
   case 0xD:
    iie.txt80=0x80;
    break;
   case 0xE:
    iie.mostxt=0x00;
    break;
   case 0xF:
    iie.mostxt=0x80;
    break;
  }
  ttyiie();
  return;
 }

 if ((a==0xC0E8)&&(fast6)) fast6on=0;
 if ((a==0xC0E9)&&(fast6)) fast6on=1;

 for (t=7; t>=(iie.iiemode?1:0); t--)
 {
  if (slot[t].dll)
  {
   o=slot[t].override_wr(a, v);
   if (!o) return;
  }
 }
 
 if ((iie.iiemode)&&((a&0xFFF0)==0xC080))
 {
  iielc(a&0xF); 
  return;  
 }
 
 if (a<0x0200)
 {
  if ((iie.iiemode)&&(iie.lcaux)) ram[65536+a]=v; else ram[a]=v;
  return;
 }
 
 if ((iie.iiemode)&&(iie.store80))
 {
  if (iie.page2)
  {
   if ((a>1023)&&(a<2048))               {ram[65536+a]=v; ttywrite(65536+a,v); return;}
   if ((a>8192)&&(a<16384)&&(iie.hires)) {ram[65536+a]=v; ttywrite(65536+a,v); return;}
  }
  else
  if (iie.page2)
  {
   if ((a>1023)&&(a<2048))               {ram[a]=v; ttywrite(a,v); return;}
   if ((a>8192)&&(a<16384)&&(iie.hires)) {ram[a]=v; ttywrite(a,v); return;}
  }
 }
 else
 {
  if ((a>1023)&&(a<2048))  
  {
   if ((iie.iiemode)&&(iie.wraux)) 
   {
    ram[65536+a]=v;
    ttywrite(65536+a,v); 
    return;
   }
   else 
   {
    ram[a]=v;
    ttywrite(a,v);
    return;
   }
  }
  if ((a>8191)&&(a<16384)) 
  {
   if ((iie.iiemode)&&(iie.wraux)) 
   {
    ram[65536+a]=v;
    ttywrite(65536+a,v); 
    return;
   }
   else 
   {
    ram[a]=v;
    ttywrite(a,v);
    return;
   }
  }
 }
 if ((a>2047)&&(a<3072))   
 {
  if ((iie.iiemode)&&(iie.wraux)) 
  {
   ram[65536+a]=v;
   ttywrite(65536+a,v); 
   return;
  }
  else 
  {
   ram[a]=v;
   ttywrite(a,v);
   return;
  }
 }
 if ((a>16383)&&(a<24576)) 
 {
  if ((iie.iiemode)&&(iie.wraux)) 
  {
   ram[65536+a]=v;
   ttywrite(65536+a,v); 
   return;
  }
  else 
  {
   ram[a]=v;
   ttywrite(a,v);
   return;
  }
 }
 
 if ((a<49152)&&(iie.iiemode)&&(iie.wraux))
 {
  ram[65536+a]=v;
  return;
 }
 if (a<49152)             
 {
  ram[a]=v; 
  return;
 }
 
 if ((a&0xFF80)==0xC080)
 {
  int s;
  s=a&0x0070;
  s>>=4;
  if (slot[s].dll)
  {
   slot[s].write16(a,v);
   return;
  }
 }
 if (((a&0xF800)==0xC000)&&(a>=0xC100))
 {
  int s;
  s=a&0x0700;
  s>>=8;
  holdc8=s;
  if (slot[s].dll)
  {
   slot[s].write256(a,v);
   return;
  }
 }

 if (((a&0xF800)==0xC800)&&holdc8)
 {
  if ((iie.iiemode)&&(iie.cxoverride)) return;
  if (!slot[holdc8].dll) return;
  slot[holdc8].write2048(a-0xC800,v);
  if (a==0xCFFF) holdc8=0;
  return;
 }

 if (a>=0xD000)
 {
  if ((iie.iiemode)&&(iie.lcwrite))
  {
   if (a<0xE000)
   {
    if (iie.lcalt)
    {
     if (iie.lcaux) {ram[65536+(a-0x1000)]=v; return;}
     
     ram[a-0x1000]=v;
    }
   }
   if (iie.lcaux) {ram[65536+a]=v; return;}
   ram[a]=v;
  }
  return;
 }
 
 if (iie.iiemode)
 {
  if (a==0xC05E) {iie.dblgrx=0x80; if (!iie.iou) iie.an3=0x00; return;}
  if (a==0xC05F) {iie.dblgrx=0x00; if (!iie.iou) iie.an3=0x80; return;}
  if (a==0xC07E) {iie.iou=0x00; return;}
  if (a==0xC07F) {iie.iou=0x80; return;}
 }
 
 if ((a&0xFFF0)==0xC000)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC010)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC030)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC050)  {ttyout(a,v); return;}
 
}

/* User level scanline; XXX might not be in synch with the video plugin */
uint16_t ulscanline;

uint8_t cpuread(uint16_t a)
{
 int t, o;
 
 if ((a==0xC0E8)&&(fast6)) fast6on=0;
 if ((a==0xC0E9)&&(fast6)) fast6on=1;

 for (t=7; t>=(iie.iiemode?1:0); t--)
 {
  if (slot[t].dll)
  {
   o=slot[t].override_rd(a);
   if (o>=0) 
   {
    return o;
   }
  }
 }

 if ((iie.iiemode)&&((a&0xFFF0)==0xC080)) 
 {
  iielc(a&0xF); 
  return 0;
 }

 if (a<0x0200)
 {
  if ((iie.iiemode)&&(iie.lcaux)) return ram[65536+a];
  return ram[a];
 }

 /* previously we used a ttyread() to get the info, but we shadow it here,
    so why waste the function call? */
 if ((iie.iiemode)&&(iie.store80))
 {
  if (iie.page2)
  {
   if ((a>1023)&&(a<2048))               return ram[65536+a];
   if ((a>8191)&&(a<16384)&&(iie.hires)) return ram[65536+a];
  }
  else
  {
   if ((a>1023)&&(a<2048))               return ram[a];
   if ((a>8191)&&(a<16384)&&(iie.hires)) return ram[a];
  }
 }
 
 if ((iie.iiemode)&&(iie.rdaux)&&(a<49152)) return ram[65536+a];
 if (a<49152)                               return ram[a];
 
 if ((a&0xFF80)==0xC080)
 {
  int s;
  s=a&0x0070;
  s>>=4;
  if (slot[s].dll)
  {
   return slot[s].read16(a);
  }
  else return 0;
 }
 
 if (((a&0xF800)==0xC000)&&(a>=0xC100))
 {
  int s;
  
  s=a&0x0700;
  s>>=8;

  holdc8=s;
  if ((iie.iiemode)&&(iie.cxoverride)) return rom[a&0x0FFF];
  if (s==3)
   if (iie.iiemode) return rom[a&0x0FFF];

  if (slot[s].dll)
  {
   return slot[s].read256(a);
  }
  return rand()&0x80; /* well, it does the job */
 }

 if (((a&0xF800)==0xC800)&&(iie.iiemode)&&(iie.cxoverride)) return rom[a-0xC000];

 if (((a&0xF800)==0xC800)&&holdc8)
 {
  uint8_t c;
  
  if (holdc8==3)
   if (iie.iiemode) return rom[a&0x0FFF];

  if (!slot[holdc8].dll) return 255;
  c=slot[holdc8].read2048(a-0xC800);
  if (a==0xCFFF) holdc8=0;
  return c;
 }

 if (a>=0xD000)
 {
  if (iie.iiemode)
  {
   if (iie.lcread)
   {
    if (a<0xE000)
    {
     if (iie.lcalt)
     {
      if (iie.lcaux) return ram[65536+(a-0x1000)];
      return ram[a-0x1000];
     }
    }
    if (iie.lcaux) return ram[65536+a];
    return ram[a];
   }
   return rom[a-0xC000];
  }
  return rom[a-0xD000];
 }
 
 if ((a&0xFFF0)==0xC000)  return ttyin(a);
 if ((a&0xFFF0)==0xC010)
 {
  if (iie.iiemode)
  {
   uint8_t k;
   
   if (!(a&0x0F)) k=ttyin(a)&0x7F; else k=ttyin(0)&0x7F;
   switch (a&0x0F)
   {
    case 0x0:
     break;
    case 0x1:
     k|=iie.lcalt;
     break;
    case 0x2:
     k|=iie.lcread;
     break;
    case 0x3:
     k|=iie.rdaux;
     break;
    case 0x4:
     k|=iie.wraux;
     break;
    case 0x5:
     k|=iie.cxoverride;
     break;
    case 0x6:
     k|=iie.lcaux;
     break;
    case 0x7:
     k|=iie.c3override;
     break;
    case 0x8:
     k|=iie.store80;
     break;
    case 0x9:
     k|=(ulscanline<192)?0x80:0x00;
     break;
    case 0xA:
     k|=(iie.nogx?0x80:0x00);
     break;
    case 0xB:
     k|=(iie.split?0x80:0x00);
     break;
    case 0xC:
     k|=(iie.page2?0x80:0x00);
     break;
    case 0xD:
     k|=(iie.hires?0x80:0x00);
     break;
    case 0xE:
     k|=iie.mostxt;
     break;
    case 0xF:
     k|=iie.txt80;
     break;
   }
   return k;
  }
  return ttyin(a);
 }

 if (iie.iiemode)
 {
  if (a==0xC05E) {iie.dblgrx=0x80; if (!iie.iou) iie.an3=0x00; return 0;}
  if (a==0xC05F) {iie.dblgrx=0x00; if (!iie.iou) iie.an3=0x80; return 0;}

  if ((a==0xC061)||(a==0xC069)) return iie.alt[0];
  if ((a==0xC062)||(a==0xC06A)) return iie.alt[1];
  
  if (a==0xC07E) {return iie.iou;}
  if (a==0xC07F) {return iie.dblgrx;}
 }
 
 if ((a&0xFFF0)==0xC030)  return ttyin(a);
 if ((a&0xFFF0)==0xC050)  return ttyin(a);

 return 0;
}

static int die;

void reset_all(void)
{
 int t;
 
 fast6on=0;
 
 ulscanline=0;
 for (t=7; t>=0; t--)
 {
  if (slot[t].dll)
   slot[t].reset();
 }
 if (iie.iiemode)
 {
  iie.rdaux=iie.wraux=iie.lcaux=0;
  iie.cxoverride=0;
  iie.nogx=iie.split=iie.page2=iie.hires=0;
  iie.store80=iie.dblgrx=iie.txt80=iie.mostxt=0;
  iie.c3override=iie.an3=0x00;
 }
 cpureset();
}

uint8_t MHz;

LARGE_INTEGER currenttime,throttlerate;
long long wantedtime;

/* Imprecise */
static void throttle (void)
{
 QueryPerformanceCounter(&currenttime);
 QueryPerformanceFrequency(&throttlerate);
 while (currenttime.QuadPart<wantedtime)
 {
  QueryPerformanceCounter(&currenttime);
  SwitchToThread();
 }
 wantedtime=currenttime.QuadPart+(throttlerate.QuadPart/(2*15720*MHz));
}

uint8_t ipdump;

/*****************************************************************************/

/*****************************************************************************
 * CPU trace - so I can try to figure out what's wrong with the emulation    *
 *****************************************************************************/

typedef enum {Nothing, Byte, Word, Relative} NumberSize;

typedef struct
{
 char Description[20];
 NumberSize Method;
}
Opcode;

Opcode Ops[256]=
{
 "BRK",Nothing,                 /* $00 */
 "ORA ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "TSB $%02X",Byte,
 "ORA $%02X",Byte,
 "ASL $%02X",Byte,
 "???",Nothing,
 "PHP",Nothing,                 /* $08 */
 "ORA #$%02X",Byte,
 "ASL",Nothing,
 "???",Nothing,
 "TSB $%04X",Word,
 "ORA $%04X",Word,
 "ASL $%04X",Word,
 "???",Nothing,
 "BPL $%04X",Relative,          /* $10 */
 "ORA ($%02X),Y",Byte,
 "ORA ($%02X)",Byte,
 "???",Nothing,
 "TRB $%02X",Byte,
 "ORA $%02X,X",Byte,
 "ASL $%02X,X",Byte,
 "???",Nothing,
 "CLC",Nothing,                 /* $18 */
 "ORA $%04X,Y",Word,
 "INC",Nothing,
 "???",Nothing,
 "TRB $%04X",Word,
 "ORA $%04X,X",Word,
 "ASL $%04X,X",Word,
 "???",Nothing,
 "JSR $%04X",Word,              /* $20 */
 "AND ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "BIT $%02X",Byte,
 "AND $%02X",Byte,
 "ROL $%02X",Byte,
 "???",Nothing,
 "PLP",Nothing,                 /* $28 */
 "AND #$%02X",Byte,
 "ROL",Byte,
 "???",Nothing,
 "BIT $%04X",Word,
 "AND $%04X",Word,
 "ROL $%04X",Word,
 "???",Nothing,
 "BMI $%04X",Relative,          /* $30 */
 "AND ($%02X),Y",Byte,
 "AND ($%02X)",Byte,
 "???",Nothing,
 "BIT $%02X,X",Byte,
 "AND $%02X,X",Byte,
 "ROL $%02X,X",Byte,
 "???",Nothing,
 "SEC",Nothing,                 /* $38 */
 "AND $%04X,Y",Word,
 "DEC",Nothing,
 "???",Nothing,
 "BIT $%04X,X",Word,
 "AND $%04X,X",Word,
 "ROL $%04X,X",Word,
 "???",Nothing,
 "RTI",Nothing,                 /* $40 */
 "EOR ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "???",Nothing,
 "EOR $%02X",Byte,
 "LSR $%02X",Byte,
 "???",Nothing,
 "PHA",Nothing,                 /* $48 */
 "EOR #$%02X",Byte,
 "LSR",Nothing,
 "???",Nothing,
 "JMP $%04X",Word,
 "EOR $%04X",Word,
 "LSR $%04X",Word,
 "???",Nothing,
 "BVC $%04X",Relative,          /* $50 */
 "EOR ($%02X),Y",Byte,
 "EOR ($%02X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "EOR $%02X,X",Byte,
 "LSR $%02X,X",Byte,
 "???",Nothing,
 "CLI",Nothing,                 /* $58 */
 "EOR $%04X,Y",Word,
 "PHY",Nothing,
 "???",Nothing,
 "???",Nothing,
 "EOR $%04X,X",Word,
 "LSR $%04X,X",Word,
 "???",Nothing,
 "RTS",Nothing,                 /* $60 */
 "ADC ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "STZ $%02X",Byte,
 "ADC $%02X",Byte,
 "ROR $%02X",Byte,
 "???",Nothing,
 "PLA",Nothing,                 /* $68 */
 "ADC #$%02X",Byte,
 "ROR",Nothing,
 "???",Nothing,
 "JMP ($%04X)",Word,
 "ADC $%04X",Word,
 "ROR $%04X",Word,
 "???",Nothing,
 "BVS $%04X",Relative,          /* $70 */
 "ADC ($%02X),Y",Byte,
 "ADC ($%02X)",Byte,
 "???",Nothing,
 "STZ $%02X,X",Byte,
 "ADC $%02X,X",Byte,
 "ROR $%02X,X",Byte,
 "???",Nothing,
 "SEI",Nothing,                 /* $78 */
 "ADC $%04X,Y",Word,
 "PLY",Nothing,
 "???",Nothing,
 "JMP ($%04X,X)",Word,
 "ADC $%04X,X",Word,
 "ROR $%04X,X",Word,
 "???",Nothing,
 "BRA $%04X",Relative,          /* $80 */
 "STA ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "STY $%02X",Byte,
 "STA $%02X",Byte,
 "STX $%02X",Byte,
 "???",Nothing,
 "DEY",Nothing,                 /* $88 */
 "BIT #$%02X",Byte,
 "TXA",Nothing,
 "???",Nothing,
 "STY $%04X",Word,
 "STA $%04X",Word,
 "STX $%04X",Word,
 "???",Nothing,
 "BCC $%04X",Relative,          /* $90 */
 "STA ($%02X),Y",Byte,
 "STA ($%02X)",Byte,
 "???",Nothing,
 "STY $%02X,X",Byte,
 "STA $%02X,X",Byte,
 "STX $%02X,Y",Byte,
 "???",Nothing,
 "TYA",Nothing,                 /* $98 */
 "STA $%04X,Y",Word,
 "TXS",Nothing,
 "???",Nothing,
 "STZ $%04X",Word,
 "STA $%04X,X",Word,
 "STZ $%04X,X",Word,
 "???",Nothing,
 "LDY #$%02X",Byte,             /* $A0 */
 "LDA ($%02X,X)",Byte,
 "LDX #$%02X",Byte,
 "???",Nothing,
 "LDY $%02X",Byte,
 "LDA $%02X",Byte,
 "LDX $%02X",Byte,
 "???",Nothing,
 "TAY",Nothing,                 /* $A8 */
 "LDA #$%02X",Byte,
 "TAX",Nothing,
 "???",Nothing,
 "LDY $%04X",Word,
 "LDA $%04X",Word,
 "LDX $%04X",Word,
 "???",Nothing,
 "BCS $%04X",Relative,          /* $B0 */
 "LDA ($%02X),Y",Byte,
 "LDA ($%02X)",Byte,
 "???",Nothing,
 "LDY $%02X,X",Byte,
 "LDA $%02X,X",Byte,
 "LDX $%02X,Y",Byte,
 "???",Nothing,
 "CLV",Nothing,                 /* $B8 */
 "LDA $%04X,Y",Word,
 "TSX",Nothing,
 "???",Nothing,
 "LDY $%04X,X",Word,
 "LDA $%04X,X",Word,
 "LDX $%04X,Y",Word,
 "???",Nothing,
 "CPY #$%02X",Byte,             /* $C0 */
 "CMP ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "CPY $%02X",Byte,
 "CMP $%02X",Byte,
 "DEC $%02X",Byte,
 "???",Nothing,
 "INY",Nothing,                 /* $C8 */
 "CMP #$%02X",Byte,
 "DEX",Nothing,
 "???",Nothing,
 "CPY $%04X",Word,
 "CMP $%04X",Word,
 "DEC $%04X",Word,
 "???",Nothing,
 "BNE $%04X",Relative,          /* $D0 */
 "CMP ($%02X),Y",Byte,
 "CMP ($%02X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "CMP $%02X,X",Byte,
 "DEC $%02X,X",Byte,
 "???",Nothing,
 "CLD",Nothing,                 /* $D8 */
 "CMP $%04X,Y",Word,
 "PHX",Nothing,
 "???",Nothing,
 "???",Nothing,
 "CMP $%04X,X",Word,
 "DEC $%04X,X",Word,
 "???",Nothing,
 "CPX #$%02X",Byte,             /* $E0 */
 "SBC ($%02X,X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "CPX $%02X",Byte,
 "SBC $%02X",Byte,
 "INC $%02X",Byte,
 "???",Nothing,
 "INX",Nothing,                 /* $E8 */
 "SBC #$%02X",Byte,
 "NOP",Nothing,
 "???",Nothing,
 "CPX $%04X",Word,
 "SBC $%04X",Word,
 "INC $%04X",Word,
 "???",Nothing,
 "BEQ $%04X",Relative,          /* $F0 */
 "SBC ($%02X),Y",Byte,
 "SBC ($%02X)",Byte,
 "???",Nothing,
 "???",Nothing,
 "SBC $%02X,X",Byte,
 "INC $%02X,X",Byte,
 "???",Nothing,
 "SED",Nothing,                 /* $F8 */
 "SBC $%04X,Y",Word,
 "PLX",Nothing,
 "???",Nothing,
 "???",Nothing,
 "SBC $%04X,X",Word,
 "INC $%04X,X",Word,
 "???",Nothing
};

int dasm(char *output, unsigned short int address)
{
 /* Addresses used by branch ops are relative to address+2, signed char */
 unsigned int num;
 signed char Rel;
 unsigned int p1,p2;
 unsigned int op;
 char Desc[128];

 /* We must keep the compiler from promoting signed values, so often we
    explicitly cast to unsigned char. */
 op=cpuread(address);

 if (Ops[op].Method==Nothing) /* Just the opcode */
 {
  sprintf (output,"%04X:  %02X            %s",address,op,Ops[op].Description);
  return address+1;
 }
 if (Ops[op].Method==Byte) /* Two byte opcode */
 {
  sprintf (Desc,Ops[op].Description,(unsigned char)(p1=cpuread(address+1)));
  sprintf (output,"%04X:  %02X%02X          %s",address,op,(unsigned char)p1,Desc);
  return address+2;
 }
 if (Ops[op].Method==Word) /* Three byte opcode */
 {
  p1=(unsigned char)cpuread(address+1);
  p2=(unsigned char)cpuread(address+2);
  num=p1+(p2<<8); /* Construct low-endian number */
  sprintf (Desc,Ops[op].Description,num);
  sprintf (output,"%04X:  %02X%02X%02X        %s",address,op,p1,p2,Desc);
  return address+3;
 }
 if (Ops[op].Method==Relative) /* Two byte opcode */
 {
  /* This is a little tricky... The relative address is a signed char. */
  Rel=cpuread(address+1);
  num=(address+2);
  num+=Rel; /* this CAN subtract; remember a+(-b) is the same as a-b */
  sprintf (Desc,Ops[op].Description,(unsigned int)num);
  sprintf (output,"%04X:  %02X%02X          %s",address,op,(unsigned char)Rel,Desc);
  return address+2;
 }
 /* error : should never reach this point but placate compilers */
 return -1;
}

/*****************************************************************************/

void dropline(void)
{
 uint16_t ip;
 char buf[80];
 
 ip=cpugetpc();
 printf("%04X- A=%02X X=%02X Y=%02X S=%02X P=%02X\n",ip,cpugeta(),cpugetx(),cpugety(),cpugets(),cpugetp());
 dasm(buf,ip);
 printf("%s\n",buf);
}

int8_t cpuloop(uint32_t cycles)
{
 int x;
 
 if (ipdump) dropline();
 
 ulscanline++;
 if (!fast6on) throttle();
 while (ulscanline>261)
 {
  ulscanline-=262;
 }
 
 for (x=iie.iiemode?1:0;x<8;x++)
 {
  if (slot[x].dll) slot[x].loop(cpugetcycles());
 }
 x=ttyloop();
 if (x==6502)
 {
  ipdump=!ipdump;
  cpusetrate(ipdump?1:65);
  printf ("--\n");
 }
 if (x==-1) return 3;
 if (x==1) reset_all();
 if ((x>=10)&&(x<18))
 {
  if (slot[x-10].dll)
  {
   slot[x-10].config();
  }
 }
 return 0;
}

int slotassign (int slotno, char *module)
{
 int e;
 
 if ((slotno<0)||(slotno>7))
 {
  fprintf (stderr, ESTR_BADSLOT, slotno);
  return -1;
 }
 
 slot[slotno].dll=LoadLibrary(module);
 if (!(slot[slotno].dll))
 {
  fprintf(stderr, ESTR_OPEN, module);
  return -1;
 }

 if ((!(slot[slotno].mod_type=(void*)GetProcAddress(slot[slotno].dll,MOD_TYPE)))||
     (!(slot[slotno].mod_name=(void*)GetProcAddress(slot[slotno].dll,MOD_NAME)))||
     (!(slot[slotno].mod_pref=(void*)GetProcAddress(slot[slotno].dll,MOD_PREF)))||
     (!(slot[slotno].mod_vers=(void*)GetProcAddress(slot[slotno].dll,MOD_VERS)))||
     (!(slot[slotno].mod_copr=(void*)GetProcAddress(slot[slotno].dll,MOD_COPR)))) notamod(module);
 
 if ((*(slot[slotno].mod_type))!=0x544F4C53)
 {
  FreeLibrary(slot[slotno].dll);
  slot[slotno].dll=0;
  fprintf (stderr, ESTR_ISNOT, module);
  fprintf (stderr, ESTR_NOTSLOT);
  return -1;
 }
 
 SFUNCTINO(slotno, preload_name, "preload_name");
 SFUNCTINO(slotno, preload_size, "preload_size");
 SFUNCTINO(slotno, preload_ptr,  "preload_ptr");

 SFUNCTINO(slotno, write16, "write16");
 SFUNCTINO(slotno, write256, "write256");
 SFUNCTINO(slotno, write2048, "write2048");

 SFUNCTINO(slotno, read16, "read16");
 SFUNCTINO(slotno, read256, "read256");
 SFUNCTINO(slotno, read2048, "read2048");

 SFUNCTINO(slotno, override_rd, "override_rd");
 SFUNCTINO(slotno, override_wr, "override_wr");
 
 SFUNCTINO(slotno, reset, "reset");
 SFUNCTINO(slotno, loop, "loop");
 SFUNCTINO(slotno, init, "init");
 SFUNCTINO(slotno, deinit, "deinit");

 SFUNCTINO(slotno, config, "config");
 
 if (*(slot[slotno].preload_ptr))
 {
  if (bload(*(slot[slotno].preload_name),*(slot[slotno].preload_size),*(slot[slotno].preload_ptr),0))
  {
   FreeLibrary(slot[slotno].dll);
   slot[slotno].dll=0;
   return -1;
  }
 }
 
 e=slot[slotno].init();
 if (e==6502) /* plugin needs dma */
 {
  SFUNCTINO(slotno, dmainit, "dmainit");
  slot[slotno].dmainit(cpuwrite, cpuread);
 }
 else if (e)
 {
  fprintf (stderr, ESTR_INITMOD, module);
  FreeLibrary(slot[slotno].dll);
  slot[slotno].dll=0;
  return -1;
 }
 
 printf (ESTR_SLOT, slotno);
 printf (ESTR_IS,*(slot[slotno].mod_name),*(slot[slotno].mod_vers),*(slot[slotno].mod_copr));
 return 0;
}

int WINAPI WinMain(HINSTANCE ghinstance, HINSTANCE dummy, LPSTR cmdline, int nCmdShow)
{
 char inifile[1024];
 char modbuf[256], slotstr[6];
 char ttymod[256], cpumod[256];
 int s,l;
 int gotrom;
 
 hinstance=ghinstance;
 
 gotrom=0;
 
 getcwd(inifile,1010);
 strcat(inifile,"\\modapple.ini");
 
 for (l=0; l<8; l++) slot[l].dll=0;
 
 printf ("%s\n", verstr);
 
 l=*ttymod=*cpumod=0;
 
 l=GetPrivateProfileString(INI_SYSTEM, INI_CONSOLE, 0, modbuf, 256, inifile);
 if (l)
 {
  strcpy(ttymod, modbuf);
 }
 
 l=GetPrivateProfileString(INI_SYSTEM, INI_CPU, 0, modbuf, 256, inifile);
 if (l)
 {
  strcpy(cpumod, modbuf);
 }
 
 for (s=0; s<8; s++)
 {
  sprintf (slotstr, "Slot%d", s);
  l=GetPrivateProfileString(INI_SLOTCARDS, slotstr, 0, modbuf, 256, inifile);
  if (l)
  {
   if (slotassign(s,modbuf))
    fprintf (stderr, ESTR_LINK, modbuf, s);
  }
 }
 
 if (!*ttymod)
 {
  fprintf (stderr, ESTR_NOTTY);
  return -1;
 }
 
 if (!*cpumod)
 {
  fprintf (stderr, ESTR_NOCPU);
  return -1;
 }
 
 mod_tty=0;
 mod_cpu=0;
 
 mod_tty=LoadLibrary(ttymod);
 if (!mod_tty)
 {
  fprintf(stderr, ESTR_OPEN, ttymod);
  return -1;
 }
 if ((!(tty_mod_type=(void*)GetProcAddress(mod_tty,MOD_TYPE)))||
     (!(tty_mod_name=(void*)GetProcAddress(mod_tty,MOD_NAME)))||
     (!(tty_mod_pref=(void*)GetProcAddress(mod_tty,MOD_PREF)))||
     (!(tty_mod_vers=(void*)GetProcAddress(mod_tty,MOD_VERS)))||
     (!(tty_mod_copr=(void*)GetProcAddress(mod_tty,MOD_COPR)))) notamod(ttymod);
 
 if ((*tty_mod_type)!=0x50595454)
 {
  canlibs();
  fprintf (stderr, ESTR_ISNOT, ttymod);
  fprintf (stderr, ESTR_NOTTTY);
  return -1;
 }
 
 TFUNCTINO(ttykey,"ttykey");
 TFUNCTINO(ttywrite,"ttywrite");
 TFUNCTINO(ttyiie,"ttyiie");
 TFUNCTINO(ttyout,"ttyout");
 TFUNCTINO(ttyin,"ttyin");
 TFUNCTINO(ttydeinit,"ttydeinit");
 TFUNCTINO(ttyinit,"ttyinit");
 TFUNCTINO(ttyloop,"ttyloop");
 
 mod_cpu=LoadLibrary(cpumod);
 if (!mod_cpu)
 {
  fprintf(stderr, ESTR_OPEN, cpumod);
  return -1;
 }
 if ((!(cpu_mod_type=(void*)GetProcAddress(mod_cpu,MOD_TYPE)))||
     (!(cpu_mod_name=(void*)GetProcAddress(mod_cpu,MOD_NAME)))||
     (!(cpu_mod_pref=(void*)GetProcAddress(mod_cpu,MOD_PREF)))||
     (!(cpu_mod_vers=(void*)GetProcAddress(mod_cpu,MOD_VERS)))||
     (!(cpu_mod_copr=(void*)GetProcAddress(mod_cpu,MOD_COPR)))) notamod(cpumod);
 
 if ((*cpu_mod_type)!=0x32303536)
 {
  canlibs();
  fprintf (stderr, ESTR_ISNOT, cpumod);
  fprintf (stderr, ESTR_NOTCPU);
  return -1;
 }
 
 CFUNCTINO(cpureset,"cpureset");
 CFUNCTINO(cpurun,"cpurun");
 CFUNCTINO(cpuinit,"cpuinit");
 CFUNCTINO(cpuseta,"cpuseta");
 CFUNCTINO(cpusetx,"cpusetx");
 CFUNCTINO(cpusety,"cpusety");
 CFUNCTINO(cpusets,"cpusets");
 CFUNCTINO(cpusetp,"cpusetp");
 CFUNCTINO(cpusetpc,"cpusetpc");
 CFUNCTINO(cpusetrate,"cpusetrate");
 CFUNCTINO(cpugeta,"cpugeta");
 CFUNCTINO(cpugetx,"cpugetx");
 CFUNCTINO(cpugety,"cpugety");
 CFUNCTINO(cpugets,"cpugets");
 CFUNCTINO(cpugetp,"cpugetp");
 CFUNCTINO(cpugetpc,"cpugetpc");
 CFUNCTINO(cpugetrate,"cpugetrate");
 CFUNCTINO(cpugetcycles,"cpugetcycles");

 l=GetPrivateProfileString(INI_SYSTEM, INI_BIOS, INI_DEF_ROM, modbuf, 256, inifile);

 iie.iiemode=0;
 
 if (!(bload(modbuf,12288,rom,1))) 
  gotrom=1;
 else
  if (!(bload(modbuf,16384,rom,0)))
  {
   iie.iiemode=1;
   gotrom=1;
  }
 if (!gotrom) fprintf (stderr, ESTR_NOBIOS);

 printf (ESTR_CPU);
 printf (ESTR_IS, *cpu_mod_name, *cpu_mod_vers, *cpu_mod_copr);
 printf (ESTR_TTY);
 printf (ESTR_IS, *tty_mod_name, *tty_mod_vers, *tty_mod_copr);

 holdc8=0;
 
 cpuinit(cpuwrite, cpuread, cpuloop);

 cpureset();
 cpusetrate(65);
// cpusetrate(1);
 
 MHz=1;
 wantedtime=0;
 fast6on=0;
 fast6=1;
 ipdump=0;
 
 *ttykey=nCmdShow;
 ttyinit(&iie);
 reset_all();
 cpurun();
 ttydeinit();
 for (l=7; l>=0; l--) if (slot[l].dll) slot[l].deinit();
 
 canlibs();
 
 return 0;
}
